/* 
 * File:   jsgfParser.h
 * Author: mschoepf
 *
 * Created on June 2, 2013, 2:35 PM
 */

#ifndef JSGFPARSER_H
#define	JSGFPARSER_H
#include <boost/variant.hpp>
#include <boost/variant/static_visitor.hpp>
#include <boost/variant/apply_visitor.hpp>
#include <boost/variant/recursive_wrapper.hpp>
#include <boost/spirit/include/qi.hpp>
#include <boost/spirit/include/qi_lit.hpp>
#include <boost/spirit/include/qi_operator.hpp>
#include <boost/spirit/include/qi_optional.hpp>
#include <boost/spirit/include/phoenix_core.hpp>
#include <boost/spirit/include/phoenix_operator.hpp>
#include <boost/spirit/include/phoenix_object.hpp>
#include <boost/spirit/include/phoenix_fusion.hpp>
#include <boost/spirit/include/phoenix_stl.hpp>
#include <boost/spirit/include/phoenix_bind.hpp>
#include <boost/fusion/include/adapt_struct.hpp>
#include <boost/fusion/include/io.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/make_shared.hpp>
#include <boost/bind.hpp>
#include <iostream>
#include <string>
#include <map>

namespace qi = boost::spirit::qi;
namespace px = boost::phoenix;
namespace ascii = boost::spirit::ascii;

namespace jsgfParserTypes {

    enum SymbolType {
        terminal,
        nonterminal,
    };
    
    struct SymbolT {
        std::string name;
        int type;
    };

    struct TermLeafT {
        std::vector<SymbolT> symbols;
    };

    struct TermT;
    
    typedef boost::variant<
        boost::recursive_wrapper<TermT>
        , TermLeafT
    > TermWrapperT;

    struct TermT
    {
        bool optional;
        bool orTerm;
        std::vector<TermWrapperT> children;        
    };
    struct RuleT {
        std::string name;
        TermT RHS;
    };
    struct GrammarT {
        std::string name;
        std::vector<RuleT> rules;
    };
    typedef qi::rule<std::string::iterator,std::vector<std::string>(), ascii::space_type> ruleT;
    typedef ruleT* ruleTPtr;
    typedef std::map<std::string, jsgfParserTypes::TermT > ruleNameToTermT;
    typedef std::map<std::string,std::vector<ruleTPtr>::size_type> nameToRuleT;
    
}
BOOST_FUSION_ADAPT_STRUCT(
        jsgfParserTypes::SymbolT,
        (std::string, name)
        (int, type)
        )

BOOST_FUSION_ADAPT_STRUCT(
        jsgfParserTypes::TermLeafT,
        (std::vector<jsgfParserTypes::SymbolT>, symbols)
        )
BOOST_FUSION_ADAPT_STRUCT(
        jsgfParserTypes::TermT,
        (bool, optional)
        (bool, orTerm)
        (std::vector<jsgfParserTypes::TermWrapperT>,children)        
)

BOOST_FUSION_ADAPT_STRUCT(
        jsgfParserTypes::RuleT,
        (std::string, name)
        (jsgfParserTypes::TermT, RHS)
        )
BOOST_FUSION_ADAPT_STRUCT(
        jsgfParserTypes::GrammarT,
        (std::string,name)
        (std::vector<jsgfParserTypes::RuleT>,rules)
)


namespace jsgfParserTypes {
    TermT
    makePTerm(TermT &t);

    TermT
    makeOptTerm(TermT &t);


    TermT
    makeOrTerm(TermT &t);

    void
    printNonterminal(const std::string s1, const std::string s2);


    std::vector<std::string>
    concat2stdvectors(const std::vector<std::string> &v1, const std::vector<std::string> &v2);

    std::vector<std::string>
    nonTerminalInserter(std::vector<std::string> &v, const std::string &s1);

    void
    printStringVector(std::vector<std::string> &v);

    int const tabsize = 4;

    void tab(int indent);

    class TermTPrinter 
    {
    public:
        TermTPrinter(int indent = 0)
          : indent(indent)
        {
        }
        void operator()(TermT const& t) const;
        int indent;
    };

    class TermWrapperTPrinter : public boost::static_visitor<> {
    public:

        TermWrapperTPrinter(int indent = 0)
        : indent(indent) {
        }
        void operator()(TermT const& t) const;
        void operator()(TermLeafT const& tl) const;
        int indent;
    };

    
    class TermTExtractor  {
    public:
        TermTExtractor (std::vector<std::string> &names) : myNames(names){}
        void operator()(TermT const& t) const;
        std::vector<std::string> &myNames;
    };
    
    class TermWrapperExtractor : public boost::static_visitor<> {
    public:
        TermWrapperExtractor (std::vector<std::string> &names) : myNames(names){}
        void operator ()(TermT const& t) const;
        
        void operator()(TermLeafT const& tl) const;        
        std::vector<std::string> &myNames;
    };    

    std::string trim(const std::string& str,
            const std::string& whitespace = " \t"); 
    class TermSanitizer {
    public:
        TermSanitizer (){}
        void operator() (TermT & t) const;
    };

    class giveChildrenAndDelete : public boost::static_visitor<std::vector<TermWrapperT> >
    {
        public:
        std::vector<TermWrapperT> operator()(TermT &t) const;
        std::vector<TermWrapperT> operator()(TermLeafT & tl) const;
    };
    
    class TermWrapperSanitizer : public boost::static_visitor<> {
    public:
        TermWrapperSanitizer(bool &isOrTerm) : isOrTerm(isOrTerm) {
        }
        void operator() (TermT &t) const; 
        void operator() (TermLeafT & tl) const;
        bool &isOrTerm;
    };
        
    class TermTRuleGenerator {
    public:
        TermTRuleGenerator (std::vector<ruleTPtr > &rules, nameToRuleT &namesToRules, std::string nonterminal) : myRules(rules), myNamesToRules(namesToRules), myNonterminal(nonterminal){}
        void operator()(TermT & t) const;
        std::vector<ruleTPtr> &myRules;
        nameToRuleT &myNamesToRules;
        std::string myNonterminal;
    };
    
    class TermWrapperRuleGenerator : public boost::static_visitor<> {
    public:
        TermWrapperRuleGenerator (std::vector<ruleTPtr> &rules, nameToRuleT & namesToRules, std::string nonterminal,bool optional) : myRules (rules), myNamesToRules(namesToRules), myNonterminal(nonterminal), optional(optional){}
        void operator ()(TermT & t) const;        
        void operator()(TermLeafT & tl) const;        
        std::vector<ruleTPtr> &myRules;
        nameToRuleT &myNamesToRules; 
        std::string myNonterminal;
        bool optional;
    };
}


template <typename Iterator>
struct jsgfParserRuleGenerator : qi::grammar<Iterator, jsgfParserTypes::GrammarT(), ascii::space_type> {
public:

    jsgfParserRuleGenerator() : jsgfParserRuleGenerator::base_type(jsgfParser, "ruleParser") {
        jsgfIdentifier = qi::lexeme["#JSGF V1.0;"];
        grammarIdentifier %= qi::lexeme["grammar"] > qi::lexeme [(qi::char_("a-zA-Z_0-9")>> *(qi::char_("a-zA-Z_0-9")))] > ";";
        identifier = qi::lexeme[(qi::char_("a-zA-Z_0-9") >> *(qi::char_("a-zA-Z_0-9 ")))];
        terminal = identifier[px::at_c<0>(qi::_val) = qi::_1] > qi::eps[px::at_c<1>(qi::_val) = px::val(0)];
        nonterminal = qi::lit('<') > identifier[px::at_c<0>(qi::_val) = qi::_1] > qi::lit('>') > qi::eps[px::at_c<1>(qi::_val) = px::val(1)];
        symbolParser = (terminal | nonterminal);
        symbolList = +(symbolParser)[px::push_back(px::at_c<0>(qi::_val),qi::_1)];
        symbolListWrapper = symbolList [qi::_val = qi::_1];
        termList = +(symbolList[px::push_back(px::at_c<2>(qi::_val),qi::_1)] | parenteseTerm[px::push_back(px::at_c<2>(qi::_val),qi::_1)] | optionalTerm[px::push_back(px::at_c<2>(qi::_val), qi::_1)] | orTerm[px::push_back(px::at_c<2>(qi::_val),qi::_1)]); //>>                 
        parenteseTerm = qi::lit('(') > termList[qi::_val = px::bind(&jsgfParserTypes::makePTerm,qi::_1)] > qi::lit(')');
        optionalTerm = qi::lit('[') > termList[qi::_val = px::bind(&jsgfParserTypes::makeOptTerm,qi::_1)] > qi::lit(']');
        orTerm = qi::lit('|') > termList[qi::_val = px::bind(&jsgfParserTypes::makeOrTerm,qi::_1)];// > *(termList[px::push_back(px::at_c<1>(qi::_val),qi::_1)]));        
        ruleParser %= (qi::eps || "public") >> qi::lit('<') > identifier/*[px::at_c<0>(qi::_val) = qi::_1]*/ > qi::lit('>') > qi::lit('=') > termList /*[px::at_c<1>(qi::_val),qi::_1]*/ > qi::lit(';');
        jsgfParser = jsgfIdentifier >> grammarIdentifier[px::at_c<0>(qi::_val),qi::_1] >> *ruleParser[px::push_back(px::at_c<1>(qi::_val),qi::_1)];
        
        
        identifier.name("Identifier");
        terminal.name("Terminal");
        nonterminal.name("Non-Terminal");
        symbolParser.name("SymbolParser");
        symbolList.name("SymbolList");
        symbolListWrapper.name("SymbolListWrapper");
        symbolListOptional.name("List of Optional Symbols");
        termParser.name("TermParser");
        termList.name("TermList");
        ruleParser.name("RuleParser");
        orTerm.name("orTerm");
        parenteseTerm.name("parenteseTerm");
        optionalTerm.name("optionalTerm");
#ifdef JSGFPARSER_DEBUG
        qi::debug(ruleParser);
        qi::debug(termList);
        qi::debug(symbolList);
        qi::debug(orTerm);
#endif
        qi::on_error<qi::fail>
        (
             jsgfParser
            , std::cout
                << px::val("Error! Expecting ")
                << qi::_4                               // what failed?
                << px::val(" here: \"")
                << px::construct<std::string>(qi::_3, qi::_2)   // iterators to error-pos, end
                << px::val("\"")
                << std::endl
        );
    }
    qi::rule<Iterator, ascii::space_type> jsgfIdentifier;
    qi::rule<Iterator, std::string(), ascii::space_type> grammarIdentifier;
    qi::rule<Iterator, std::string(), ascii::space_type> identifier;
    qi::rule<Iterator, jsgfParserTypes::SymbolT(), ascii::space_type> terminal;
    qi::rule<Iterator, jsgfParserTypes::SymbolT(), ascii::space_type> nonterminal;
    qi::rule<Iterator, jsgfParserTypes::SymbolT(), ascii::space_type> symbolParser;
    qi::rule<Iterator, jsgfParserTypes::TermLeafT(), ascii::space_type> symbolList;
    qi::rule<Iterator, jsgfParserTypes::TermWrapperT(), ascii::space_type> symbolListWrapper;
    qi::rule<Iterator, jsgfParserTypes::TermT(), ascii::space_type> parenteseTerm;
    qi::rule<Iterator, jsgfParserTypes::TermT(), ascii::space_type> optionalTerm;
    qi::rule<Iterator, jsgfParserTypes::TermT(), ascii::space_type> orTerm;
    qi::rule<Iterator, jsgfParserTypes::TermLeafT(), ascii::space_type> symbolListOptional;
    qi::rule<Iterator, jsgfParserTypes::TermWrapperT(), ascii::space_type> termParser;
    qi::rule<Iterator, jsgfParserTypes::TermT(), ascii::space_type> termList;
    qi::rule<Iterator, jsgfParserTypes::TermT(), ascii::space_type> termListList;
    qi::rule<Iterator, jsgfParserTypes::RuleT(), ascii::space_type> ruleParser;
    qi::rule<Iterator, jsgfParserTypes::GrammarT(),ascii::space_type> jsgfParser;
    jsgfParserTypes::ruleNameToTermT rulesMap;
    std::vector<jsgfParserTypes::ruleTPtr > myRuleVector;

    bool sanitizeCheckGrammar(jsgfParserTypes::GrammarT &grammar, bool verbose = false) {
        jsgfParserTypes::TermTPrinter myTermPrinter;
        jsgfParserTypes::TermSanitizer mySanitizer;
        for (std::vector<jsgfParserTypes::RuleT>::size_type g = 0; g < grammar.rules.size(); g++) {
            jsgfParserTypes::RuleT myRule = grammar.rules.at(g);
            rulesMap.insert(std::make_pair(myRule.name, myRule.RHS));
            if (verbose) {
                std::cout << "Parsing successfully completed: Rule name = '" << myRule.name << "'" << std::endl;
            }
            mySanitizer(myRule.RHS);
            if (verbose)
                myTermPrinter(myRule.RHS);
            grammar.rules.at(g) = myRule;
        }
        std::cout << "Map size: " << rulesMap.size() << std::endl;
        std::vector<std::string> nonterminals;
        jsgfParserTypes::TermTExtractor myTermExtractor(nonterminals);
        for (jsgfParserTypes::ruleNameToTermT::iterator i = rulesMap.begin(); i != rulesMap.end(); i++) {
            myTermExtractor((*i).second);
        }
        std::cout << "Found " << nonterminals.size() << " nonterminals" << std::endl;
        for (std::vector<std::string>::const_iterator i = nonterminals.begin(); i != nonterminals.end(); i++) {
            if (verbose)
                std::cout << "Looking for '" << *i << "' ";
            jsgfParserTypes::ruleNameToTermT::iterator pos = rulesMap.find(*i);
            if (pos != rulesMap.end()) {
                if (verbose)
                    std::cout << "found" << std::endl;
            } else {
                if (verbose)
                    std::cout << "NOT FOUND" << std::endl;
                return false;
            }
        }
        return true;
    }

    bool generateRules(jsgfParserTypes::GrammarT &grammar) {

        jsgfParserTypes::nameToRuleT myNameToRule;

        myRuleVector.clear();
        std::vector<jsgfParserTypes::ruleTPtr>::size_type increment = 0;
        for (std::vector<jsgfParserTypes::ruleTPtr>::size_type i = 0; i < grammar.rules.size(); i++) // Make an empty rule for every rule/nonterminal we read
        {
            //myRuleVector.push_back (boost::make_shared<jsgfParserTypes::ruleT>());
            myRuleVector.push_back(new jsgfParserTypes::ruleT(qi::eps));
            (*myRuleVector.back()).name(grammar.rules.at(i).name);
            myNameToRule.insert(std::make_pair(grammar.rules.at(i).name, increment));
            increment++;
        }
        increment = grammar.rules.size() - 1;
        for (std::vector<jsgfParserTypes::ruleTPtr>::size_type i = grammar.rules.size(); i > 0; i--) {
            std::vector<jsgfParserTypes::ruleTPtr> tempRuleVector = myRuleVector;
            jsgfParserTypes::TermTRuleGenerator myRuleGenerator(myRuleVector, myNameToRule, grammar.rules.at(i - 1).name);
            //std::cout << "***Making Rule " << myRuleVector.at(increment)->name() << std::endl;
            myRuleGenerator(grammar.rules.at(i - 1).RHS);
            std::string ruleResult;
            if (0 == increment)
                (*myRuleVector.at(increment)) = (*myRuleVector.back())[qi::_val = px::bind(&jsgfParserTypes::nonTerminalInserter, qi::_1, px::val(grammar.rules.at(i - 1).name))] > qi::eps;
            else
                (*myRuleVector.at(increment)) = (*myRuleVector.back()) > qi::eps;
            ruleResult = (*myRuleVector.back()).name();
            std::cout << "*** Rule " << myRuleVector.at(increment)->name() << " = " << ruleResult << std::endl;
            //std::vector<jsgfParserTypes::ruleTPtr>::size_type sizeDiff = myRuleVector.size() - tempRuleVector.size();        
            increment--;
        }
        return true;
    }

    std::vector<jsgfParserTypes::ruleTPtr> getRules() {
        return myRuleVector;
    }
    //virtual ~parserToyExample();
private:

};



#endif	/* JSGFPARSER_H */

