#include <boost/smart_ptr/make_shared.hpp>

#include "configFileParser.h"
#include "SimpleDoubleThresholdSegmenter.h"
#include "PocketSphinxInitException.h"

PocketSphinxInstanceConfig::PocketSphinxInstanceConfig() : name(PSW_DEFAULT_NAME),hmmDir(""),lmDir(""),dictDir(""),jsgfDir(""),bestpath(false),autoload(false){}

PocketSphinxGlobalConfig::PocketSphinxGlobalConfig() : startDB(SIMPLETHRESHOLD_STARTDB), uttDB(SIMPLETHRESHOLD_UTTDB), silence(SIMPLETHRESHOLD_PAUSE), timeout(SIMPLETHRESHOLD_MAXLENGTH),alsaDevice(ALSADEFAULTDEVICE) {}


void
PocketSphinxInstanceConfig::parseSection(sectionT globalSection) {
    name = globalSection.first.second;
    
    for (std::vector<pair_type>::const_iterator it = globalSection.second.begin(); it != globalSection.second.end(); it++) {
        bool haveParsedElement = false;
        
        if (boost::iequals(it->first, "hmm")) {
            hmmDir = it->second;
            haveParsedElement = true;
        }
        if (!haveParsedElement && boost::iequals(it->first, "name")) {
            name = it->second;
            haveParsedElement = true;
        }
        if (!haveParsedElement && boost::iequals(it->first, "lm")) {
            lmDir = it->second;
            haveParsedElement = true;
        }
        if (!haveParsedElement && boost::iequals(it->first, "dict")) {
            dictDir = it->second;
            haveParsedElement = true;
        }
        if (!haveParsedElement && boost::iequals(it->first, "jsgf")) {
            jsgfDir = it->second;
            haveParsedElement = true;
        }
        if (!haveParsedElement && boost::iequals(it->first, "rsbscope")) {
            rsbScope = it->second;
            haveParsedElement = true;
        }
        if (!haveParsedElement && boost::iequals(it->first, "bestpath")) {
            if (boost::iequals(it->second, "true") || boost::iequals(it->second, "on") || boost::iequals(it->second, "yes")) {
                bestpath = true;
                haveParsedElement = true;
            }
            if (boost::iequals(it->second, "false") || boost::iequals(it->second, "off") || boost::iequals(it->second, "no")) {
                bestpath = false;
                haveParsedElement = true;
            }
            if (!haveParsedElement) {
                throw PocketSphinxParseException(std::string("Invalid Argument here: ").append(it->first));
            }
        }
        if (!haveParsedElement && boost::iequals(it->first, "autoload")) {
            if (boost::iequals(it->second, "true") || boost::iequals(it->second, "on") || boost::iequals(it->second, "yes")) {
                autoload = true;
                haveParsedElement = true;
            }
            if (boost::iequals(it->second, "false") || boost::iequals(it->second, "off") || boost::iequals(it->second, "no")) {
                autoload = false;
                haveParsedElement = true;
            }
            if (!haveParsedElement) {
                throw PocketSphinxParseException(std::string("Invalid Argument here: ").append(it->first));
            }
        }
        if (!haveParsedElement)
        {
            std::cerr << "Could not map argument: " << it->first << std::endl;
            throw PocketSphinxParseException(std::string("Could not map argument: ").append(it->first));
        }
    }
}


void
PocketSphinxGlobalConfig::parseSection(sectionT globalSection) {
    //sanity check
    if (0 != globalSection.first.second.compare("")) {
        std::cerr << "Unexpected value in [main] Header" << std::endl;
        throw PocketSphinxParseException("Unexpected value in [main] Header");
    }
    for (std::vector<pair_type>::const_iterator it = globalSection.second.begin(); it != globalSection.second.end(); it++) {
        bool haveParsedElement = false;
        if (boost::iequals(it->first, "startdb")) {
            try {
                startDB = boost::lexical_cast<unsigned int>(it->second);
            } catch (boost::bad_lexical_cast const&) {
                std::cerr << "Invalid Argument here: " << it->first << std::endl;
                throw PocketSphinxParseException(std::string("Invalid Argument here: ").append(it->first));
            }
            haveParsedElement = true;
        }
        if (!haveParsedElement && boost::iequals(it->first, "uttdb")) {
            try {
                uttDB = boost::lexical_cast<unsigned int>(it->second);
            } catch (boost::bad_lexical_cast const&) {
                std::cerr << "Invalid Argument here: " << it->first << std::endl;
                throw PocketSphinxParseException(std::string("Invalid Argument here: ").append(it->first));
            }
            haveParsedElement = true;
        }
        if (!haveParsedElement && boost::iequals(it->first, "silence")) {
            try {
                silence = boost::lexical_cast<unsigned int>(it->second);
            } catch (boost::bad_lexical_cast const&) {
                std::cerr << "Invalid Argument here: " << it->first << std::endl;
                throw PocketSphinxParseException(std::string("Invalid Argument here: ").append(it->first));
            }
            haveParsedElement = true;
        }
        if (!haveParsedElement && boost::iequals(it->first, "timeout")) {
            try {
                timeout = boost::lexical_cast<unsigned int>(it->second);
            } catch (boost::bad_lexical_cast const&) {
                std::cerr << "Invalid Argument here: " << it->first << std::endl;
                throw PocketSphinxParseException(std::string("Invalid Argument here: ").append(it->first));
            }
            haveParsedElement = true;
        }
        if (!haveParsedElement && boost::iequals(it->first, "alsadevice")) {
            alsaDevice = it->second;
            haveParsedElement = true;
        }

        if (!haveParsedElement && boost::iequals(it->first, "openDeviceDirectly")) {
            if (boost::iequals(it->second, "true") || boost::iequals(it->second, "on") || boost::iequals(it->second, "yes")) {
                openDeviceDirectly = true;
                haveParsedElement = true;
            }
            if (boost::iequals(it->second, "false") || boost::iequals(it->second, "off") || boost::iequals(it->second, "no")) {
                openDeviceDirectly = false;
                haveParsedElement = true;
            }
            if (!haveParsedElement) {
                throw PocketSphinxParseException(std::string("Invalid Argument here: ").append(it->first));
            }
        }
    }
}


parseConfigFile::parseConfigFile(const std::string &filename) : filename(filename)
{    
    myConfig = boost::make_shared<PocketSphinxGlobalConfig>();
}

void
parseConfigFile::doParse() {
    std::string storage; // We will read the contents here.
    std::ifstream input(filename.c_str(), std::ifstream::in);
    key_value_sequence<std::string::iterator> keysandvalues;
    comment_skipper<std::string::iterator> skipper;
    std::vector<sectionT> parsedSections;    
    
    if (!input.is_open()) {
        std::cerr << "Could not open file: " << filename << std::endl;
    }
    
    input.unsetf(std::ios::skipws); // No white space skipping!
    std::copy(
            std::istream_iterator<char>(input),
            std::istream_iterator<char>(),
            std::back_inserter(storage));
    std::string::iterator iter = storage.begin();
    std::string::iterator end = storage.end();
    bool success = qi::phrase_parse(iter, end, keysandvalues,skipper , parsedSections); 
    if (success)
    {
        if (iter != end)
        {
            std::cout << "Parsed PARTLY successfully" << std::endl;
            std::cout << "Leftover: '" << std::string().assign(iter,end) << "'" << std::endl;
            throw PocketSphinxParseException(std::string("Parsed PARTLY successfully, Leftover: '").append(std::string().assign(iter,end)));
        }        
    }
    else
    {
        std::string::iterator some = iter + 30;
        std::string context(iter, (some > end) ? end : some);
        std::cout << "-------------------------\n";
        std::cout << "Parsing failed\n";
        std::cout << "stopped at: \": " << context << "...\"\n";
        std::cout << "-------------------------\n";
        throw PocketSphinxParseException(std::string ("Parsing failed, stopped at: ").append(context));
    }
    
    for (std::vector<sectionT>::const_iterator it=parsedSections.begin();it != parsedSections.end();it++)
    {
        if (boost::iequals(it->first.first,"main"))
        {
            myConfig->parseSection(*it);
        }
        if (boost::iequals(it->first.first,"pocketsphinxinstance"))
        {
            boost::shared_ptr<PocketSphinxInstanceConfig> myLocalConfig= boost::make_shared<PocketSphinxInstanceConfig>();
            myLocalConfig->parseSection(*it);
            myConfig->AddToPocketSphinxInstanceVector(myLocalConfig);
        }
    }
    std::vector<boost::shared_ptr<PocketSphinxInstanceConfig> > instanceConfig = myConfig->GetPocketSphinxInstanceVector();
    //I Know this could be done more efficient but we make it deep down dirty here
    for (std::vector<boost::shared_ptr<PocketSphinxInstanceConfig> >::iterator it = instanceConfig.begin();it != instanceConfig.end();it++)
    {
        for (std::vector<boost::shared_ptr<PocketSphinxInstanceConfig> >::iterator it2 = instanceConfig.begin();it2 != instanceConfig.end();it2++)
        {
            if (it != it2)
            {
                if (boost::iequals((*it)->GetRsbScope(),(*it2)->GetRsbScope()))
                {
                    std::cerr << "Found two instances with the same rsb scope: " << (*it)->GetRsbScope() << std::endl;
                    throw PocketSphinxParseException(std::string("Found two instances with the same rsb scope: ").append((*it)->GetRsbScope()));
                }
            }
        }
    }
    if (myConfig->GetStartDB() > myConfig->GetUttDB())
    {
        std::cerr << "Invalid parameter: startDb > uttDb" << std::endl;
        throw PocketSphinxParseException("Invalid parameter: startDb > uttDb");
    }
    if ((myConfig->GetStartDB() < SIMPLETHRESHOLD_STARTDB_MIN) || (myConfig->GetStartDB() > SIMPLETHRESHOLD_STARTDB_MAX))
    {
        std::cerr << "startDb out of range: " << myConfig->GetStartDB() << "[" << SIMPLETHRESHOLD_STARTDB_MIN << ".." << SIMPLETHRESHOLD_STARTDB_MAX << "]" << std::endl;
        throw PocketSphinxInitException("startDb out of range"); 
    }
    if ((myConfig->GetUttDB() < SIMPLETHRESHOLD_UTTDB_MIN) || (myConfig->GetUttDB() > SIMPLETHRESHOLD_UTTDB_MAX))
    {
        std::cerr << "uttDb out of range: " << myConfig->GetUttDB() << "[" << SIMPLETHRESHOLD_UTTDB_MIN << ".." << SIMPLETHRESHOLD_UTTDB_MAX << "]" << std::endl;
        throw PocketSphinxInitException("uttDb out of range"); 
    }
    if ((myConfig->GetSilence() < SIMPLETHRESHOLD_PAUSE_MIN) || (myConfig->GetSilence() > SIMPLETHRESHOLD_PAUSE_MAX))
    {
        std::cerr << "silence out of range: " << myConfig->GetSilence() << "[" << SIMPLETHRESHOLD_PAUSE_MIN << ".." << SIMPLETHRESHOLD_PAUSE_MAX << "]" << std::endl;
        throw PocketSphinxInitException("silence Db out of range"); 
    }
    if ((myConfig->GetTimeout() < SIMPLETHRESHOLD_MAXLENGTH_MIN) || (myConfig->GetTimeout() > SIMPLETHRESHOLD_MAXLENGTH_MAX))
    {
        std::cerr << "timeout out of range: " << myConfig->GetTimeout() << "[" << SIMPLETHRESHOLD_MAXLENGTH_MIN << ".." << SIMPLETHRESHOLD_MAXLENGTH_MAX << "]" << std::endl;
        throw PocketSphinxInitException("timeout out of range"); 
    }
}

boost::shared_ptr<PocketSphinxGlobalConfig>
parseConfigFile::getConfig() {
    return myConfig;
}